package ipeirotis.gal.scripts;

import java.util.HashMap;
import java.util.TreeSet;
import java.util.Vector;

 
import ipeirotis.utils.Utils;

public class DawidSkene {

	public static void main(String[] args) {

		String inputfile = "";
		String correctfile = "";
		String costfile = "";
		int iterations = 0;
		
		if (args.length!=4) {
			System.out.println("Usage: java -jar getanotherlabel.jar <inputfile> <correctfile> <costfile> <iterations>");
			System.out.println("");
			System.out.println("Example: java -jar getanotherlabel.jar data\\unlabeled.txt data\\labeled.txt  data\\costs.txt 10");
			System.out.println("");
			System.out.println("The <inputfile> is a tab-separated text file.");
			System.out.println("Each line has the form <workerid><tab><objectid><tab><assigned_label>");
			System.out.println("and records the label that the given worker gave to that object");
			System.out.println("");
			System.out.println("The <correctfile> is a tab-separated text file.");
			System.out.println("Each line has the form <workerid><tab><objectid><tab><assigned_label>");
			System.out.println("and records the correct labels for whatever objects we have them.");
			System.out.println("The workerid is ignored for the <correctfile>");
			System.out.println("");
			System.out.println("The <costfile> is a tab-separated text file.");
			System.out.println("Each line has the form <from_class><tab><to_class><tab><classification_cost>");
			System.out.println("and records the classification cost of classifying an object that belongs to the `from_class` into the `to_class`.");
			System.out.println("");
			System.out.println("<iterations> is the number of times to run the algorithm. Even a value of 1 works well.");
			System.exit(-1); 
		} else {
			inputfile = args[0];
			correctfile  = args[1];
			costfile  = args[2];
			iterations = Integer.parseInt(args[3]);
		}
		
		String[] lines_input = inputfile.split("\n");
		Vector<Labeling> labelings = loadLabels(lines_input);
		
		String[] lines_correct = correctfile.split("\n");
		Vector<Labeling> correct = loadLabels(lines_correct);
		
		String[] lines_cost = costfile.split("\n");
		Vector<Labeling> costs = loadLabels(lines_cost);
		
		
		DawidSkene ds = new DawidSkene(labelings, correct, costs);
		
		
		// Save the majority vote before the D&S estimation
		HashMap<String,String> prior_voting = ds.getMajorityVote();
		Utils.writeFile( ds.printVote(), "data/pre-majority-vote.txt");
		
		
		ds.estimate(iterations);
		ds.updateAnnotatorCosts();

		// Save the estimated cost for each worker
		Utils.writeFile(ds.printAnnotatorCostsSummary(), "data/worker-costs.txt");
		
		// Save the error rates for the workers
		Utils.writeFile(ds.printAllWorkerScores(), "data/worker-error-rates.txt");
		
		// Save the probability that an object belongs to each class
		Utils.writeFile(ds.printObjectClassProbabilities(0.0), "data/object-probabilities.txt");
		
		// Save the majority vote after the D&S estimation
		HashMap<String,String> posterior_voting = ds.getMajorityVote();
		Utils.writeFile(ds.printVote(), "data/post-majority-vote.txt");
		
		// Save the probability that an object belongs to each class
		Utils.writeFile(ds.printDiffVote(prior_voting, posterior_voting), "data/differences-pre-post-majority-vote.txt");

	}

	public String printDiffVote(HashMap<String, String> prior_voting, HashMap<String, String> posterior_voting) {
		StringBuffer sb = new StringBuffer(); 
		
		for (String obj:(new TreeSet<String>(prior_voting.keySet()))) {
			String prior_vote = prior_voting.get(obj);
			String posterior_vote = posterior_voting.get(obj);
			
			if (prior_vote.equals(posterior_vote)) {
				sb.append("SAME\t"+obj+"\t"+prior_vote);
			} else {
				sb.append("DIFF\t"+obj+"\t"+prior_vote+"->"+posterior_vote);
			}
			sb.append("\n");
		}
		return sb.toString();
	}

	public String printVote() {
		StringBuffer sb = new StringBuffer(); 
		// la map ↓↓ contient les résultats traités ------------------------------------------------------------------------
		HashMap<String, String> vote = getMajorityVote();
		
		for (String obj:(new TreeSet<String>(vote.keySet()))) {
			String majority_vote = vote.get(obj);
			sb.append(obj+"\t"+majority_vote+"\n");
		}
		return sb.toString();
	}
	
	public static Vector<Labeling> loadLabels(String[] lines) {
		

		Vector<Labeling> labelings = new Vector<Labeling>();
		for (String line : lines) {
			// System.out.println(line);
			String[] entries = line.split("\t");

			// We need 3 entries per line. If the annotator has not provided an
			// entry, ignore.
			// In the future, we may need to treat the empty submission as a
			// separate class.
			if (entries.length == 3) {
				Labeling lb = new Labeling(entries[0], entries[1], entries[2]);
				labelings.add(lb);
			}
		}
		return labelings;
	}

	// Maps annotators, object, and classes to integers
	HashMap<String, Integer> annotators;
	HashMap<String, Integer> contributions;
	HashMap<String, Integer> objects;
	HashMap<String, Integer> classes;
	
	// Given the ID's retrieve the names
	HashMap<Integer, String> annotators_names;
	HashMap<Integer, String> objects_names;
	HashMap<Integer, String> classes_names;

	// Expected cost per annotator
	HashMap<Integer, Double> annotators_cost_naive;
	HashMap<Integer, Double> annotators_cost_adjusted;
	HashMap<Integer, Double> annotators_cost_minimized;
	
	// We have I objects
	int I;

	// We have J possible labels per object
	int J;

	// We have K annotators
	int K;

	// The labels given to each object from each annotator
	// label[k][i] is the label assigned by annotator k to object i
	// --->>>> We assign -1 if it has not been labeled!!!!
	int label[][];
	
	
	// The correct labels for each object
	// correct[i] is the correct label for object i
	// --->>>> We assign -1 if we do not know the correct label
	int correct[];

	// Error rate (confusion matrix) for each annotator
	// pi[k][j][l] is the probability that annotator k, classifies an item from
	// category j to category l
	double[][][] pi;
	
	// Cost matrix for various errors
	// cost[i][j] is the cost of classifying an item from the correct category i to the 
	double[][] cost;

	// Priors for the different classes
	double[] prior;

	// The probabilities of different labels for each object
	// T[oid][l] is the probability that the object oid belong to class l
	double T[][];
	// Tdemon[oid][l] is the normalizing constant 
	double Tdenom[][];
	// Tcount[oid] is the number of labels assigned to object oid
	double Tcount[];

	public DawidSkene(Vector<Labeling> labels, Vector<Labeling> correct_labels, Vector<Labeling> classification_cost) {

		// Assign IDs to the labelers, objects, and classes
		this.annotators = new HashMap<String, Integer>();
		this.classes = new HashMap<String, Integer>();
		this.objects = new HashMap<String, Integer>();
		this.contributions = new HashMap<String, Integer>();
		
		// Assign IDs to the labelers, objects, and classes
		this.annotators_names = new HashMap<Integer, String>();
		this.classes_names = new HashMap<Integer, String>();
		this.objects_names = new HashMap<Integer, String>();

		int aid = 0; // Annotator IDs
		int cid = 0; // Class IDs
		int oid = 0; // Object IDs
		for (Labeling l : labels) {
			String annotatorid = l.annotatorid;
			if (!annotators.containsKey(annotatorid)) {
				//System.out.println("Annotator:" + annotatorid + "=" + aid);
				annotators.put(annotatorid, aid);
				annotators_names.put(aid,annotatorid);
				contributions.put(annotatorid, 0);
				aid++;
			}

			String objectid = l.objectid;
			if (!objects.containsKey(objectid)) {
				//System.out.println("Object:" + objectid + "=" + oid);
				objects.put(objectid, oid);
				objects_names.put(oid, objectid);
				oid++;
			}

			String labelid = l.label;
			if (!classes.containsKey(labelid)) {
				//System.out.println("Label:" + labelid + "=" + cid);
				classes.put(labelid, cid);
				classes_names.put(cid,labelid);
				cid++;
			}
		}
		
		if (correct_labels != null) {
			for (Labeling l : correct_labels) {
	
				String objectid = l.objectid;
				if (!objects.containsKey(objectid)) {
					//System.out.println("Object:" + objectid + "=" + oid);
					objects.put(objectid, oid);
					objects_names.put(oid, objectid);
					oid++;
				}
	
				String labelid = l.label;
				if (!classes.containsKey(labelid)) {
					//System.out.println("Label:" + labelid + "=" + cid);
					classes.put(labelid, cid);
					classes_names.put(cid,labelid);
					cid++;
				}
			}
		}
		

		this.K = annotators.keySet().size();
		this.I = objects.keySet().size();
		this.J = classes.keySet().size();

		this.pi = new double[K][J][J];
		initializeErrorRates();
		
		this.prior = new double[J];
		initializePriors();
		
		// Initialize labels to -1 (i.e., not labeled by annotator)
		this.label = new int[K][I];
		for (int k = 0; k < K; k++) {
			for (int i = 0; i < I; i++) {
				label[k][i] = -1;
			}
		}

		// Initialize correct labels to -1 (i.e., we do not know the correct answers)
		this.correct = new int[I];
			for (int i = 0; i < I; i++) {
				this.correct[i] = -1;
			}
		
		// Load the labelings into the label array
		for (Labeling l : labels) {
			int labelerid = annotators.get(l.annotatorid);
			int objectid = objects.get(l.objectid);
			int classid = classes.get(l.label);
			
			int curr = contributions.get(l.annotatorid);
			contributions.put(l.annotatorid, curr+1);

			this.label[labelerid][objectid] = classid;
		}

		// Load the correct labels the correct array
		if (correct_labels != null) {
			for (Labeling l : correct_labels) {
				int objectid = objects.get(l.objectid);
				int classid = classes.get(l.label);
	
				this.correct[objectid] = classid;
			}
		}

		
		// Initialize the class probability assignments
		this.T = new double[I][J];
		initializeObjectClassProbabilities();
		
		// Initialize the cost array 
		this.cost = new double[J][J];
		initializeCosts();
		if (classification_cost != null) {
			// TODO: Using the Labeling class to load cost 
			// is ugly. Just create a proper class, ensuring
			// that all costs are loaded properly etc.
			for (Labeling c : classification_cost) {
				
				// The first column is the "from/correct" class
				int from = classes.get(c.annotatorid);
				
				// The second column is the "to/classified" class
				int to = classes.get(c.objectid);
				
				// The second column is the cost of classifying
				// an object from the "correct" to the "classified" class
				double decision_cost = Double.parseDouble(c.label);

				this.cost[from][to]= decision_cost;
			}
		}
		
	}
	
	private void initializeCosts() {
		
		for (int i=0; i<J; i++) {
			for (int j=0; j<J; j++) {
				if (i==j) {
					cost[i][j]=0;
				} else {
					cost[i][j]=1;
				}
				
			}
		}
		

	}
	
	public void updateAnnotatorCosts() {
		annotators_cost_naive = new HashMap<Integer, Double>();
		annotators_cost_adjusted = new HashMap<Integer, Double>();
		annotators_cost_minimized  = new HashMap<Integer, Double>();
		
		for (int k=0; k<K; k++) {
			annotators_cost_naive.put(k, getAnnotatorCostNaive(k));
			annotators_cost_adjusted.put(k, getAnnotatorCostAdjusted(k));
			annotators_cost_minimized.put(k, getAnnotatorCostMinimized(k));
		}
	}
	

	/**
	 * Estimates the cost for annotator k without attempting corrections of labels
	 * 
	 * @param k
	 * @return The expected cost of misclassifications of annotator k
	 */
	public double getAnnotatorCostNaive(int k) {
		double c = 0;
		double s=0; 
			for (int i = 0; i < J; i++) {
				for (int j = 0; j < J; j++) {
					if (pi[k][i][j]>0) {
						c += pi[k][i][j]*cost[i][j]*prior[i];
						s += prior[i]*cost[i][j];
					}
				}
			}
		if (s>0) return c/s;
		return 0;
	}
	
	
	/**
	 * Estimates the cost for annotator k after adjusting class estimates using the error rates of the annotator
	 * 
	 * @param k
	 * @return The expected cost of misclassifications of annotator k
	 */
	public double getAnnotatorCostMinimized(int k) {
		
		
		double c = 0;
		
		// The annotator will give one of the J possible labels for each example
		//
		// We estimate first how often the annotator will give label l
		// based on the "true priors" for each class and the error rates of annotator k
		
		double[] annotator_prior = new double[J];

		// Let's go over each of the classes and see how often this "true" class j 
		// gets misclassified as l
		// TODO: Create this at the initialization time. Just count how many times this
		// annotator classifies an object as l
		for (int l = 0; l < J; l++) {
			for (int j = 0; j < J; j++) {
				if (pi[k][j][l]>-0.1) {
						annotator_prior[l] += prior[j]*pi[k][j][l];
				}
			}
		}
		
		// We now know the frequency with which we will see a label "assigned_label" from annotator k
		// Each of this "hard" labels from the annotator k will corresponds to a corrected
		// "soft" label
		for (int assigned_label = 0; assigned_label < J; assigned_label++) {
			// Let's find the soft label that corresponds to assigned_label
			double[] soft = new double[J];
			for (int i = 0; i < J; i++) {
				if (pi[k][i][assigned_label]>-0.1 && annotator_prior[assigned_label]>0) {
					soft[i] = pi[k][i][assigned_label]*prior[i]/annotator_prior[assigned_label];
				}
			}
			
			// And add the cost of this label, weighted with the prior of seeing this label.
			c += getMinSoftLabelCost(soft)*annotator_prior[assigned_label];
		}
		return c/getMinSpammerCost();
	}
	
	/**
	 * Estimates the cost for annotator k after adjusting class estimates using the error rates of the annotator
	 * 
	 * @param k
	 * @return The expected cost of misclassifications of annotator k
	 */
	public double getAnnotatorCostAdjusted(int k) {
		
		
		double c = 0;
		
		// The annotator will give one of the J possible labels for each example
		//
		// We estimate first how often the annotator will give label l
		// based on the "true priors" for each class and the error rates of annotator k
		
		double[] annotator_prior = new double[J];

		// Let's go over each of the classes and see how often this "true" class j 
		// gets misclassified as l
		// TODO: Create this at the initialization time. Just count how many times this
		// annotator classifies an object as l
		for (int l = 0; l < J; l++) {
			for (int j = 0; j < J; j++) {
				if (pi[k][j][l]>-0.1) {
						annotator_prior[l] += prior[j]*pi[k][j][l];
				}
			}
		}
		
		// We now know the frequency with which we will see a label "assigned_label" from annotator k
		// Each of this "hard" labels from the annotator k will corresponds to a corrected
		// "soft" label
		for (int assigned_label = 0; assigned_label < J; assigned_label++) {
			// Let's find the soft label that corresponds to assigned_label
			double[] soft = new double[J];
			for (int i = 0; i < J; i++) {
				if (pi[k][i][assigned_label]>-0.1 && annotator_prior[assigned_label]>0) {
					soft[i] = pi[k][i][assigned_label]*prior[i]/annotator_prior[assigned_label];
				}
			}
			
			// And add the cost of this label, weighted with the prior of seeing this label.
			c += getSoftLabelCost(soft)*annotator_prior[assigned_label];
		}
		return c/getSpammerCost();
	}
	
	/**
	 * Gets as input a "soft label" (i.e., a distribution of probabilities over classes)
	 * and returns the expected cost of this soft label.
	 * 
	 * @param p
	 * @return The expected cost of this soft label
	 */
	private double getSoftLabelCost(double[] p) {
		double c = 0;
		
		for (int i = 0; i < J; i++) {
			for (int j = 0; j < J; j++) {
				c += p[i]*p[j]*cost[i][j];
			}
		}
		
		return c;
	}
	
	/**
	 * Gets as input a "soft label" (i.e., a distribution of probabilities over classes)
	 * and returns the expected cost of this soft label.
	 * 
	 * @param p
	 * @return The expected cost of this soft label
	 */
	private double getMinSoftLabelCost(double[] p) {

		double min_cost = Double.MAX_VALUE;
		for (int i = 0; i < J; i++) {
			// So, with probability p[i] it belongs to class i
			
			// What is the expected cost in this case?
			double costfor_i = 0;
			for (int j = 0; j < J; j++) {
				costfor_i += p[j]*cost[i][j];

			}
			
			if (costfor_i<min_cost) {
				min_cost = costfor_i;
			}
			
		}
		
		return min_cost;
	}
	
	/**
	 * Returns the cost of a "spammer" worker, who assigns completely random labels.
	 * 
	 * @return The expected cost of a spammer worker
	 */
	private double getSpammerCost() {
		return getSoftLabelCost(prior);
	}
	
	/**
	 * Returns the cost of a "spammer" worker, who assigns completely random labels.
	 * 
	 * @return The expected cost of a spammer worker
	 */
	private double getMinSpammerCost() {
		return getMinSoftLabelCost(prior);
	}

	public void estimate(int iterations) {
		for (int i = 0; i < iterations; i++) {
			updateObjectClassProbabilities();
			updatePriors();
			updateAnnotatorErrorRates();
		}
	}

	private void initializeErrorRates() {
		// Set error rates to be not-so perfect, to allow for errors
		for (int lid = 0; lid < K; lid++) {
			for (int j = 0; j < J; j++) {
				for (int k = 0; k < J; k++) {
					if (j == k) {
						pi[lid][j][k] = 0.9;
					} else {
						pi[lid][j][k] = 0.1 / (J - 1);
					}
				}
			}
		}
	}

	private void initializeObjectClassProbabilities() {

		for (int oid = 0; oid < I; oid++) {
			
			// First check if we already have the correct label
			// for this object. If yes, then give the appropriate
			// value to the T[oid][j] cells, and continue 
			if (correct[oid] != -1) {
				for (int j = 0; j < J; j++) {
					if (correct[oid] == j) {
						T[oid][j] = 1.0;
					} else {
						T[oid][j] = 0.0;
					}
				}
				continue;
			}
			
			// Count how many times oid has been labeled
			// as belonging to each class, and total.
			int[] cnt = new int[J];
			int total = 0;
			for (int k = 0; k < K; k++) {
				if (label[k][oid] != -1) {
					int given_label = label[k][oid];
					cnt[given_label]++;
					total++;
				}
			}

			// Count how many times oid has been labeled as class j
			for (int j = 0; j < J; j++) {
				T[oid][j] = 1.0 * cnt[j] / total;
			}
		}
	}

	private void initializePriors() {

		// Set priors to be uniform
		for (int j = 0; j < J; j++) {
			prior[j] = 1.0 / J;
		}
	}

	public String printAnnotatorCostsSummary() {
		StringBuffer sb = new StringBuffer();
		
		for (int k = 0; k < K; k++) {
			String annotator_name = annotators_names.get(k);
			Double cost_naive = annotators_cost_naive.get(k);
			Double cost_adj = annotators_cost_adjusted.get(k);
			Double cost_min = annotators_cost_minimized.get(k);

			Integer contribution = contributions.get(annotator_name);
			sb.append(annotator_name+"\t"+Utils.round(100*cost_naive, 2) + "%\t"+Math.round(100*(1-cost_adj)) + "%\t"+Math.round(100*(1-cost_min)) + "%\t" + contribution +"\n");
		}
		return sb.toString();
	}
	

	public String printAllWorkerScores() {
		StringBuffer sb = new StringBuffer();
		
		for (int k=0; k<K; k++) {
			sb.append(printWorkerScore(k));
		}
		return sb.toString();
	}
	
	
	/*
	 * Like printAnnotatorCosts, but doesn't print cost_naive and cost_adj
	 */
	public String printWorkerScore(int workerid) {
		StringBuffer sb = new StringBuffer();
		
		String annotator_name = annotators_names.get(workerid);
		Double cost_naive = annotators_cost_naive.get(workerid);
		Double cost_adj = annotators_cost_adjusted.get(workerid);
		Double cost_min = annotators_cost_minimized.get(workerid);		// score des turker ------------------------------------------------------
		
		Integer contribution = contributions.get(annotator_name);

		sb.append("Worker: "+ annotator_name + "\n");
		sb.append("Error Rate: "+ Utils.round(100*cost_naive, 2) + "%\n");
		sb.append("Quality_Expected: " + Math.round(100*(1-cost_adj)) + "%\n");
		sb.append("Quality_Optimized: " + Math.round(100*(1-cost_min)) + "%\n");
		sb.append("Number of Annotations: "  + contribution + "\n");
		sb.append("Confusion Matrix: \n");
		for (int i = 0; i < J; i++) {
			for (int j = 0; j < J; j++) {
				String correct_name = classes_names.get(i);
				String assigned_name = classes_names.get(j);
				sb.append("P[" + correct_name + "->" + assigned_name + "]=" + ( (pi[workerid][i][j]>-0.1)? Utils.round(100*pi[workerid][i][j], 3) :"----" )+ "%\t");
			}
			sb.append("\n");
		}
		sb.append("\n");

		return sb.toString();
	}

	/**
	 * Prints the objects that have probability distributions with entropy higher than the given threshold
	 * 
	 * @param entropy_threshold
	 */
	public String printObjectClassProbabilities(double entropy_threshold) {
		
		StringBuffer sb = new StringBuffer();
		for (int i = 0; i < I; i++) {
			double entropy = Utils.entropy(T[i]);
			if (entropy<entropy_threshold) continue;
			
			String object_name = objects_names.get(i);
			sb.append(object_name + "\t");
			for (int j = 0; j < J; j++) {
				String class_name = classes_names.get(j);
				sb.append("Pr[" + class_name + "]=" + T[i][j] + "\t");
			}
			sb.append("\n");
		}
		
		return sb.toString();

	}

	public String printPriors() {
		StringBuffer sb = new StringBuffer();
		for (int j = 0; j < J; j++) {
			String class_name = classes_names.get(j);
			sb.append("Prior[" + class_name + "]=" + prior[j]+"\n");
		}
		return sb.toString();
	}

	public void updateAnnotatorErrorRates() {
		for (int lid = 0; lid < K; lid++) {
			double[][] confusion_matrix = new double[J][J];
			for (int j = 0; j < J; j++) {
				for (int l = 0; l < J; l++) {
					confusion_matrix[j][l]=0; 
				}
			}
			// Scan all objects and change the confusion matrix for each annotator
			// using
			// the class probability for each object
			for (int oid = 0; oid < I; oid++) {
				int l = label[lid][oid];
				
				// If the worker has not annotated the object, 
				// we do not use it in estimating the error rate
				if (l == -1)
					continue;
				
				// We now get the classification of the object
				// based on the votes of all the other workers
				double[] T_lid = getObjectClassProbabilities(oid, lid);
				
				// If nobody else has labeled the object oid,
				// we have no information
				if (T_lid == null)
					continue;
				
				for (int j = 0; j < J; j++) {
					double d= T_lid[j];
					//double d= T[oid][j];
					//System.out.printf("d %d %d: %s\n",oid,j,Double.toString(d));
					
					confusion_matrix[j][l] += d;
				}
			}
			
			// Compute the total number of J cases
			double[] marginal = new double[J];
			for (int j = 0; j < J; j++) {
				for (int n = 0; n < J; n++) {
					marginal[j] += confusion_matrix[j][n];
				}
			}
			
			// Evaluate the values of the probabilistic error rates using the
			// confusion matrix
			for (int j = 0; j < J; j++) {
				for (int l = 0; l < J; l++) {
					pi[lid][j][l] = (marginal[j] > 0) ? Utils.round(1.0 * confusion_matrix[j][l] / marginal[j], 5) : -1.0;
					
					// Version with Laplace smoothing
					// pi[lid][j][l] =  Utils.round( (confusion_matrix[j][l]+1) / (marginal[j]+J), 5);
				}
			}
		}
		//System.out.print(printAnnotatorCosts2(true));
	}

	public void updateObjectClassProbabilities() {
		for (int oid = 0; oid < I; oid++) {
			
			// First check if we already have the correct label
			// for this object. If yes, then give the appropriate
			// value to the T[oid][j] cells, and continue 
			if (correct[oid] != -1) {
				// So if we **have** gold data, we get in this.
				
				// Note: In principle, the following for-loop
				// is unnecessary. The gold T values should have been
				// correctly set during the initialization phase
				// But we keep this here, just in case.
				for (int j = 0; j < J; j++) {
					if (correct[oid] == j) {
						T[oid][j] = 1.0;
					} else {
						T[oid][j] = 0.0;
					}
				}
				
				continue;
				
			}
			
			// If we do not have the correct label, then we proceed as 
			// usual with the M-phase of the EM-algorithm of Dawid&Skene
			for (int j = 0; j < J; j++) {
				// Estimate nominator for Eq 2.5 of Dawid&Skene
				double nom =  prior[j];
				// Go through all annotators
				for (int k = 0; k < K; k++) {
					// Find the label given by annotator K to the object oid
					int given_label = label[k][oid];
					if (given_label == -1)
						continue;
					// Having the annotator gave the given_label, and given how often
					// given_label is mixed with j
					double evidence_for_j = pi[k][j][given_label];
					
					// Use the evidence only if we have a real estimate for pi[k][j][l] (i.e., it is not -1)
					if ( pi[k][j][given_label] >-0.1 ) {
						nom *= evidence_for_j;
					}
				}
				
				// Estimate denominator for Eq 2.5 of Dawid&Skene
				double denom = 0;
				for (int q = 0; q < J; q++) {
					double pdenom = prior[q];
					for (int k = 0; k < K; k++) {
						int given_label = label[k][oid];
						if (given_label == -1)
							continue;
						double evidence_for_q = pi[k][q][given_label];
						
						// Use the evidence only if we have a real estimate for pi[k][j][l] (i.e., it is not -1)
						if ( pi[k][q][given_label] >-0.1 ) {
							pdenom *= evidence_for_q;
						}
					}
					denom += pdenom;
				}
				T[oid][j] = (denom > 0) ? Utils.round( nom / denom, 5) : 0.0;
				//Tdenom[oid][j] = (denom > 0) ? Utils.round( nom / denom, 5) : 0.0;

			}			
		}
	}

	public double[] getObjectClassProbabilities(int oid, int lid) {
		
			
			// First check if we already have the correct label
			// for this object. If yes, then give the appropriate
			// value to the T[oid][j] cells, and continue 
			if (correct[oid] != -1) {
				// So if we **have** gold data, we get in this.
				
				// Note: In principle, the following for-loop
				// is unnecessary. The gold T values should have been
				// correctly set during the initialization phase
				// But we keep this here, just in case.
				for (int j = 0; j < J; j++) {
					if (correct[oid] == j) {
						T[oid][j] = 1.0;
					} else {
						T[oid][j] = 0.0;
					}
				}
				
				return T[oid];
				
			}
			
	
			
			// If we do not have the correct label, then we proceed as 
			// usual with the M-phase of the EM-algorithm of Dawid&Skene
			
			// Estimate denominator for Eq 2.5 of Dawid&Skene
			double denom = 0;
			for (int q = 0; q < J; q++) {
				double pdenom = prior[q];
				for (int k = 0; k < K; k++) {
					int given_label = label[k][oid];
					if (given_label == -1)
						continue;
					if (k == lid) 
						continue;
					
					double evidence_for_q = pi[k][q][given_label];
					
					// Use the evidence only if we have a real estimate for pi[k][j][l] (i.e., it is not -1)
					if ( pi[k][q][given_label] >-0.1 ) {
						pdenom *= evidence_for_q;
					}
				}
				denom += pdenom;
			}

			// If the denominator is 0, we have no basis for creating probability distribution
			if (denom == 0) return null;
			
			// This is the object probabilities without the influence of the annotator
			double Tnew[] = new double[J];
			
			for (int j = 0; j < J; j++) {
				// Estimate nominator for Eq 2.5 of Dawid&Skene
				double nom =  prior[j];
				// Go through all annotators
				for (int k = 0; k < K; k++) {
					// Find the label given by annotator K to the object oid
					int given_label = label[k][oid];
					if (given_label == -1)
						continue;
					if (k == lid) 
						continue;
					// Having the annotator gave the given_label, and given how often
					// given_label is mixed with j
					double evidence_for_j = pi[k][j][given_label];
					
					// Use the evidence only if we have a real estimate for pi[k][j][l] (i.e., it is not -1)
					if ( pi[k][j][given_label] >-0.1 ) {
						nom *= evidence_for_j;
					}
				}
				

				Tnew[j] = Utils.round( nom / denom, 5);

			}
			return Tnew;
		
	}
	
	public void updatePriors() {
		for (int j = 0; j < J; j++) {
			double pr = 0;
			for (int oid = 0; oid < I; oid++) {
				pr += T[oid][j];
			}
			prior[j] = Utils.round(pr / I, 5);
		}
	}
	
	private int getMajorityClass(int oid) {
		double max = -1;
		int majorityclass = -1;
		
		for (int j = 0; j < J; j++) {
			if (T[oid][j]>max) {
				max = T[oid][j];
				majorityclass = j;
			} else if (T[oid][j]==max) {
				// In case of a tie, break ties based on the priors
				if (prior[j]>prior[majorityclass]) {
					max = T[oid][j];
					majorityclass = j;
				}
			}
			
			
		}
		return majorityclass;
	}
	
	public HashMap<String,String> getMajorityVote() {
		HashMap<String,String> result = new HashMap<String,String>();
		
		for (int oid=0; oid<I; oid++) {
			String object_name = objects_names.get(oid);
			int majority = getMajorityClass(oid);
			String class_name = classes_names.get(majority);
			
			result.put(object_name, class_name);
		}
		return result;
	}

	// Methods I added
	
	public HashMap <Integer,String>  getAnnotatorsName()
	{
		return this.annotators_names ;
	}
	
	public HashMap <Integer,Double> getAnnotatorsCost()
	{
		return this.annotators_cost_minimized ;
	}
	
}




