package myShell;

import java.io.BufferedReader;
import java.io.BufferedWriter;
import java.io.File;
import java.io.FileInputStream;
import java.io.FileOutputStream;
import java.io.FileWriter;
import java.io.InputStream;
import java.io.InputStreamReader;
import java.util.*;

import myDataBases.MediaBase;

import org.jdom.*;
import org.jdom.input.SAXBuilder;
import org.jdom.output.Format;
import org.jdom.output.XMLOutputter;



/**
 * Deals with several xml files, either reading or writing them.
 * 
 * <p>This class reads the configuration file <tt>config.xml</tt> to retrieve the constants used in this program, as well
 * as the axes along which comparisons must be performed and the structure of the information used (i.e : fields of the
 * <tt>hit</tt> table and their matching elements in the csv file retrieved from the Antechamber).</p>
 * 
 * <p>It also writes intermediary results in order for the splitsort to continue where it stopped as well as final results
 * (those you want if you are using this software).</p>
 * 
 * @author	Leo Perrin (perrin.leo@gmail.com)
 */
public class Interpreter {

	//
	// Fields
	//

	/**
	 * The {@link myDataBases.MediaBase} in which the names and other characteristics of the media sorted are stored.
	 */
	private MediaBase mediaBase;
	/**
	 * The number of media sorted.
	 */
	private Integer numberOfMedia;

  
	//
	// Constructors
	//

	/**
	 * Creates an Interpreter instance.
	 * 
	 * <p>This instance will use the {@link myDataBases.MediaBase} MediaBase to get (among other tings) the names of the
	 * media. It also parses previous results in order to know how much media are to be sorted this time.</p>
	 * 
	 * @param		database	The MediaBase instance to be aggregated to the Interpreter.
	 */
	public Interpreter (MediaBase database)
	{
		// media
		this.mediaBase = database;
		this.numberOfMedia = 0;
		for (Integer[] array : this.mediaBase.getAllMedia())
			this.numberOfMedia += array.length;
	}
  
	//
	// Methods
	//



// ============================== SETTINGS =========================================


	/**
	 * Reads the value of several parameters in the "../data/config.xml" file.
	 * 
	 * <p>This method reads the content of the "constants" node of the config.xml file. It should contain all the constants
	 * used by CPS, such as the path to file where media are stored the login used to connect with the database.</p>
	 * 
	 * <p>The results are stored in a map : each parameter is in the field which key is its identifier (for example,
	 * <tt>result["mediaPath"] = "../data/media.xml"</tt> (default).</p>
	 */
	public static Map <String,String> getConfig()
	{
		Element rootConfig;
		org.jdom.Document documentConfig;
	    Map <String,String> conf = new Hashtable <String,String>() ;
		SAXBuilder sxb = new SAXBuilder();
		try
		{
			documentConfig= sxb.build(new File("../data/config.xml"));
	 	    rootConfig = documentConfig.getRootElement();
		    Element constants = rootConfig.getChild("constants");
			Iterator it = constants.getChildren().iterator();
			while (it.hasNext())
			{
				Element current = (Element)it.next();
				conf.put(current.getName(), current.getValue());
			}
		}
	    catch(Exception e){ e.printStackTrace(); }
		return conf ;
	}
  
  
	/**
	 * Returns a List of String containing the axes considered during the comparisons.
	 * 
	 * <p>Returns all the children of the "axes" node of the config.xml file whose name is "axis". They correspond to the
	 * axes along which the sorting is to be performed.</p>
	 * 
	 * @return	A list of String, each String being an axis.
	 */
	public static List<String> getAxes()
	{

		Element rootConfig;
		org.jdom.Document documentConfig;
		SAXBuilder sxb = new SAXBuilder();
		List<String> listAxes = new ArrayList <String>() ;
		try
		{
			documentConfig= sxb.build(new File("../data/config.xml"));
			rootConfig = documentConfig.getRootElement();
			Element axes = rootConfig.getChild("axes");
			if (axes != null)
			{
				Iterator it = axes.getChildren().iterator();
				while (it.hasNext())
				{
					Element current = (Element)it.next();
					listAxes.add(current.getValue());
				}
			}
			else
				listAxes = null;
		}
		catch(Exception e){ e.printStackTrace(); }
		return listAxes;
	}




	/**
	 * Returns a List of String containing the names of the fields used to retrieve hard data from the HIT results file.
	 * 
	 * <p>What is returned depends on the value of the parameter <tt>fields</tt>. If it equals :
	 * <ul>
	 *		<li> "SQLfield" : it returns the fields of the SQL table HITresults in which hard results are stored</li>
	 *		<li> "CSVfield" : it returns the fields of the CSV file containing HIT results returned by CrowdFlower</li>
	 *		<li> "SQLtype"  : it returns the types of the SQL fields, i.e "INT", "VARCHAR(40)", etc.</li>
	 *		<li> anything else : returns null.</li>
	 * </ul>
	 * </p>
	 * 
	 * @param		fields The name of the field to retrieve. It must have one of these values : "SQLfield",
	 * "CSVfield" or "SQLtype"
	 * 
	 * @return		The list of the fields wanted.				
	 */
	public static List<String> getFields(String fields)
	{
		Element rootConfig;
		org.jdom.Document documentConfig;
		List<String> listFields = new ArrayList <String>() ;
		if ( (fields.equals("SQLfield")) || (fields.equals("SQLtype")) || (fields.equals("CSVfield")) )
		{
			SAXBuilder sxb = new SAXBuilder();
			try
			{
				documentConfig= sxb.build(new File("../data/config.xml"));
				rootConfig = documentConfig.getRootElement();
				Element HDfields = rootConfig.getChild("hardDataFields");
				Iterator it = HDfields.getChildren().iterator();
				while (it.hasNext())
				{
					Element current = (Element)it.next();
					String entry = current.getChild(fields).getValue() ;
					if (entry.length() > 1)
						listFields.add(entry);
				}
			}
			catch(Exception e){ e.printStackTrace(); }
		}
		else
			listFields = null;
		return listFields;
	}


	/**
	 * Checks if the <tt>config.xml</tt> file contains all the fields necessary for CPS to work.
	 * 
	 * <p>This function verifies if the configuration file contains :
	 * <ul>
	 * 		<li>mediaPath: the path to the file containing information on the media</li>
	 * 		<li>sqlBase: the name of the MySQL database used.
	 * 		<li>sqlUser: the user of the MySQL database</li>
	 * 		<li>antechamberUrl: the URL of the Antechamber, a website in charge of receiving results from CF and parse them
	 * (among other things).</li>
	 * 		<li>cfKey: the key given by CrowdFlower (it is available <a href="https://crowdflower.com/account">here</a>
	 * 			if you have an account. And you should, otherwise there is no chance for CPS to work !)</li>
	 * 		<li>cfJobNumber: the identifier of the job used. This time, it is
	 * <a href="https://crowdflower.com/jobs">here</a></li>
	 * 		<li>At least one axis along which to perform comparisons.</li>
	 * 		<li>At least the <tt>media1, media2, greater</tt> and <tt>axis</tt> fields in the <tt>hardDataFields</tt>
	 * 			node</li>
	 * </ul>
	 * </p>
	 * 
	 * @param	log		The LogWriter instance where the results are saved.
	 * 
	 * @return	<tt>true</tt> if everything is fine in the <tt>config.xml</tt> file, <tt>false</tt> otherwise.
	 */
	public static Boolean checkFile(LogWriter log)
	{
		Boolean fine = true;
		log.append("------- Checking config.xml");

		// checking the constants
		Map<String,String> configContent = getConfig() ;
		String[] necessaryConstants = { "mediaPath", "sqlBase", "sqlUser", "antechamberUrl", "cfKey", "cfJobNumber" };
		for (String constant : necessaryConstants)
			if (configContent.keySet().contains(constant))
				log.append(constant+" constant found");
			else
			{
				log.append("ERROR : " + constant + " NOT FOUND. Interrupting.");
				fine = false;
			}
		
		// checking the presence of axes
		List<String> axes = getAxes() ;
		if (axes == null)
		{
			fine = false;
			log.append("ERROR : no axis found. Interrupting.");
		}
		else
			log.append(axes.size() + " axes found.");
		
		// checking the content of the hard data fields
		List<String> fields = getFields("SQLfield");
		String[] necessaryFields = { "media1", "media2", "greater", "axis" };
		if (fields == null)
			log.append("ERROR : no 'hardDataFields' section in the config.xml. Interrupting.");
		else
			for (String field : necessaryFields)
				if (fields.contains(field))
					log.append(field +" found in the hardDataFields section");
				else
				{
					log.append("ERROR : " + field + " NOT FOUND in the hardDataFields section. Interrupting.");
					fine = false;
				}
		
		log.append("------- END check config.xml\n");
		return fine;
	}



// ============================== PREVIOUS RESULTS =====================================


	/**
	 * Returns the Integer saved in the '../data/previously/lastJob' file.
	 * 
	 * @return	The identifier of the last job used.
	 */
	public static Integer getLastJob()
	{
		Integer id = 0;
		try{
			InputStream ips = new FileInputStream("../data/previously/lastJob"); 
			InputStreamReader ipsr = new InputStreamReader(ips);
			BufferedReader br = new BufferedReader(ipsr);
			id = Integer.parseInt(br.readLine());
			br.close(); 
		}		
		catch (Exception e){ e.printStackTrace(); }
		return id;
	}
	
	
	/**
	 * Saves a job identifier in the '../data/previously/lastJob' file.
	 * 
	 * @param	id	The identifier of the job to save.
	 */
	public static void setLastJob(Integer id)
	{
		try
		{
			FileWriter fw = new FileWriter("../data/previously/lastJob", false);
			BufferedWriter output = new BufferedWriter(fw);
			output.write(id.toString());
			output.flush();
			output.close();
		} catch(Exception e) { e.printStackTrace(); }
	}


	/**
	 * Parses the xml file containing the current state of the sort along the <tt>axis</tt> axis.
	 * 
	 * <p>Reads the <tt>'../data/previously/$AXIS.xml'</tt> file and returns its content. This file should be formatted so
	 * as to contain a list of arrays of Integer.</p>
	 * 
	 * @param 	axis	The axis from which we want to retrieve the results.
	 * 
	 * @return	The state of the splitsort along this axis.
	 */
	private List<Integer[]> parsePrevious(String axis)
	{
		Element rootPrevious;
		org.jdom.Document documentPrevious;
		List<Integer[]> result = new ArrayList<Integer[]>();
		SAXBuilder sxb = new SAXBuilder();
		try
		{
			documentPrevious = sxb.build(new File("../data/previously/" + axis + ".xml"));
	 	    rootPrevious = documentPrevious.getRootElement();
			Iterator it = rootPrevious.getChildren().iterator();
			while (it.hasNext())
			{
				Element array = (Element)it.next();
				List<Integer> arrayContent = new ArrayList<Integer>();
				Iterator jt = array.getChildren().iterator();
				while (jt.hasNext())
				{
					Element entry = (Element)jt.next();
					arrayContent.add (Integer.parseInt(entry.getValue()) );
				}
				result.add(arrayContent.toArray(new Integer[0]));
			}
		}
		catch(Exception e){ e.printStackTrace(); }
		
		return result;
	}

	
	/**
	 * Returns the current state of the SplitSort along the <tt>axis</tt> axis if there is a file containing it in
	 * <tt>../data/previously</tt>, otherwise returns the result of <tt>getAllMedia()</tt>.
	 * 
	 * <p>This function checks if an <tt>axis.xml</tt> file in the <tt>../data/previously/</tt> directory exists. If so, it
	 * parses its content and returns it in a list of arrays. Each element of this list is an array waiting for a splitsort
	 * iteration (or not, if its size is too small).</p>
	 * <p>If there is no such file, it calls <tt>mediaBase.getAllMedia()</tt> in order to get a new list of the media to
	 * sort.</p>
	 * 
	 * @param	log		The LogWriter instance in which save the advancement.
	 * @param 	axis	The axis from which we want to retrieve the results.
	 * 
	 * @return	The state of the media along the <tt>axis</tt> at the end of the previous iteration or, in its absence,
	 * all media's Id.	
	 */
	public List<Integer[]> getCurrentState(LogWriter log, String axis)
	{
		List<Integer[]> lst;
		File previouslyInCPS = new File("../data/previously/");
		// checking if the file exists.
		Integer count = 0;
		for (String file : previouslyInCPS.list())
			if (file.equals(axis + ".xml"))
				break;
			else
				count ++;
		if (count == previouslyInCPS.list().length)
		{
			lst = new ArrayList<Integer[]>();
			lst = this.mediaBase.getAllMedia();
			log.append("[DONE] Reading media to sort.\n   ... Note: no previous results found for " + axis +
					": starting from the beginning.");
		}
		else
		{
			lst = parsePrevious(axis);
			log.append("[DONE] Reading media to sort.\n   ... Note: previous results found for " + axis +
					": starting where we left.");
		}
		return lst;
	}



// ============================== SAVING RESULTS =====================================
	
	
	
	/**
	 * Generates an xml file containing the current state of the splitsort.
	 * 
	 * <p>Generates a file formatted so as to store the content of a list of array. The aim is for this file to be used as
	 * a starting point in a next splitsort iteration by the <tt>getCurrentState()</tt> method.</p>
	 * 
	 * @param	xmlFile	The file in which the current state of an axis is being saved right now.
	 * @param	content	The organization of the media along a given axis during the current iteration. For instance, if
	 * there are three arrays at the end of this iteration along the <tt>axis</tt> axis, it must contain the following
	 * entry : { [2,1,3],[4],[7,6,5] }.
	 */
	public void generateIntermediaryXMLs(String xmlFile, List<Integer[]> content)
	{

		Element rootResults = new Element("results");
		org.jdom.Document documentResults = new Document(rootResults);
		
		Map<Integer, Map<String,String>> mediaCharac= this.mediaBase.getContent();
		Map<String,String> media ;
		Element charac;
		
		for (Integer[] array : content)
		{
			Element arrayElement = new Element("array");
			for (Integer mediaID : array)
			{
				Element identifier = new Element("id");
				identifier.setText(String.valueOf(mediaID));
				arrayElement.addContent(identifier);
			}
			rootResults.addContent(arrayElement);
		}
		try
		{
			XMLOutputter sortie = new XMLOutputter(Format.getPrettyFormat());
			sortie.output(documentResults, new FileOutputStream(xmlFile));
		}
		catch (Exception e){ e.printStackTrace(); }
	}

	  
	/**
	 * Transforms a rank among the other media into a real value along an axis.
	 * 
	 * <p>In this case, we assume that the media follow a uniform distribution in [-1,1], therefore the abscissa is given
	 * by: \f${ {2 \times rank} \over {numberOfMedia - 1} } - 1\f$ The first media, at the <tt>0th</tt>th rank, will have
	 * -1 as abscissa and the last one, at the <tt>(numberOfMedia - 1)</tt>th rank, will have 1.</p>
	 * 
	 * @param	rank	The rank of the media along an axis
	 * 
	 * @return	The value corresponding to this rank.
	 */
	private Double rank2value(Integer rank)
	{
		Double value =  ( (2.0*rank)/(this.numberOfMedia-1) ) - 1.0;
		return value;
	}


	/**
	 * Transforms the map firstMap into a List in order it to be used by the generateFinalXML() method.
	 * 
	 * <p>The Map of Arrays returned by the last loop of the main function has a structure that is not suited for writing
	 * with a correct formatting in the results.xml file. Therefore, this method returns a List of Map that is suited
	 * for such a writing. The structure of the input and output is explained below. 
	 * </p>
	 * 
	 * @param		sortResults 	The map to modify. Its structure is :
	 * {<tt>axis considered</tt>, [<tt>order of the media along this axis</tt>]}.
	 * 
	 * @return	A list of map corresponding to the input map of arrays. Its structure is :
	 * [ {<tt>axis</tt>,<tt>rank along this axis of the media whose ID is its rank in the list of map</tt>}... ]
	 * 
	 * @see Interpreter.#generateFinalXML(String, Map)
	 */
	private List< Map<String,Integer> > preformating(Map<String,Integer[]> sortResults)
	{
		List< Map<String,Integer> > newMap = new ArrayList< Map<String,Integer> >();
		// loop on the ID of the media considered
		for (int media=0; media<this.numberOfMedia; media++)
		{
			Map<String,Integer> mediaRanks = new Hashtable<String,Integer>();
			// loop on the label of the axis considered
			for ( String axis : sortResults.keySet() )
			{
				int rank = 0;
				// loop to find the rank of the media along the "axis" axis
				while (rank<sortResults.get(axis).length)
					if (sortResults.get(axis)[rank] == media)
						break;
					else
						rank++ ;
				mediaRanks.put(axis,rank);
			}
			newMap.add(mediaRanks);
		}
		return newMap;
	}


	/**
	 * Writes an xml file containing the position of each media along each axis as well as all their characteristics
	 * stored in the <tt>dbMedia</tt>; returns also the content of this file in a String for further treatment.
	 * 
	 * <p>The results.xml file has the following structure :</p>
	 * <p>< sorted >< media >< characteristic >a first characteristic< /characteristic >
	 * < othercharac >a second characteristic< /othercharac >< axis >abscissa along this axis< /axis >
	 * < otherAxis >abscissa along this other axis< /otherAxis >< /media >...< /sorted>
	 * </p>
	 * 
	 * @param		xmlFile 	The path where the xml file will be written.
	 * @param		sortResults The results to be written in the xml file. It must have the following structure:
	 * <tt>[ "axis considered", "order of the media along this axis"]</tt>.
	 * 
	 * @return		An XML formatted String which is identical to what is written in the file.
	 */
	public String generateFinalXML(String xmlFile, Map<String,Integer[]> sortResults)
	{
		Element rootResults = new Element("results");
		org.jdom.Document documentResults = new Document(rootResults);
		Map<Integer, Map<String,String>> mediaCharac= this.mediaBase.getContent();
		Map<String,String> media ;
		List< Map<String,Integer> > results = this.preformating(sortResults) ;
		Element charac;
		String xmlString = "A problem was encoutered!";
		for (Integer mediaID : mediaCharac.keySet())
		{
			// retrieving information on this media
			media = mediaCharac.get(mediaID);
			// creation of a new node corresponding to a new media
			Element mediaElement = new Element("media");
			Attribute identifier = new Attribute("id",String.valueOf(mediaID));
			mediaElement.setAttribute(identifier);
			for (String entry : media.keySet())
			{
				charac = new Element(entry);
				charac.setText(media.get(entry));
				mediaElement.addContent(charac);
			}
			// addition of its values along each axis
			Map<String,Integer> resultsForThisMedia = results.get(mediaID); 
			for ( Iterator<String> emotion = resultsForThisMedia.keySet().iterator(); emotion.hasNext(); )
			{
				String axis = emotion.next();
				Integer rank = resultsForThisMedia.get(axis);
				Element coordinate = new Element(axis);
				coordinate.setText(this.rank2value(rank).toString());
				mediaElement.addContent(coordinate);
			}
			rootResults.addContent(mediaElement);
		}
		// Writing the xml Document in a hard file
		try
		{
			XMLOutputter sortie = new XMLOutputter(Format.getPrettyFormat());
			sortie.output(documentResults, new FileOutputStream(xmlFile));
			xmlString = sortie.outputString(documentResults);
		}
		catch (Exception e){ e.printStackTrace(); }
		return xmlString;
	}

}
