package myShell;

// java tools
import java.io.BufferedWriter;
import java.io.FileWriter;
import java.util.*;
// CPS specific classes
import sorter.SplitSort;
import sorter.Comparator;
import myDataBases.MySQLbase;
import myDataBases.MediaBase;
import crowdUser.CrowdManager;


/**
 * The main class of the Crowd-Powered-Sort program.
 * 
 * <p>This class contains the <tt>main</tt> method that parses command line arguments to call the right function. It also
 * implements :
 * <ul>
 * 		<li>The displaying of a quick help for the user if he asks for it or if no arguments were used.</li>
 * 		<li>The creation of a sql file containing the information necessary to create the DBcps database used by this
 * program.</li>
 * </ul>
 * </p>
 * 
 * @author	Leo Perrin (perrin.leo@gmail.com)
*/
public class MainClass {
	
	//
	//	Attributes
	//


	/**
	 * A {@link LogWriter} instance used to... write a log file. Isn't it amazing ?
	 */
	public static LogWriter log;
	/**
	 * A map containing the data within the <tt>config.xml</tt> file.
	 */
	public static Map<String,String> config;
	/**
	 * The {@link myDataBases.MySQLbase} instance used to store the job results, both raw and refined.
	 */
	public static MySQLbase hitBdd;
	/**
	 * A {@link myDataBases.MediaBase} used to store the content of the <tt>media.xml</tt> file.
	 */
	public static MediaBase mediaBdd;
	/**
	 * A list of array containing all the media to sort and the advancement of the sort.
	 */
	public static List<Integer[]> allMedia;
	/**
	 * The {@link crowdUser.CrowdManager} instance supposed to be the interface with Crowdflower and the Antechamber
	 * during the splitsort.
	 */
	public static CrowdManager theBoss;
	/**
	 * This {@link Interpreter} will read the <tt>config.xml</tt> file and write the <tt>results.xml</tt> one. 
	 */
	public static Interpreter inter;
	/**
	 * The {@link sorter.Comparator} instance that will perform the SplitSort along each specified axis.
	 */
	public static Comparator comp;
	/**
	 * Is <tt>true</tt> if the sort must continue; is <tt>false</tt> when everything is over.
	 */
	public static Boolean keepGoing;
	
	//
	//	Methods
	//
	/**
	 * Main method of this program.
	 * 
	 * <p>Parses command line option to launch the right function. If no argument is specified, help is displayed using
	 * the <tt>displayHelp()</tt> method. A description of what is done depending on the command line arguments is
	 * given in the body of the <tt>displayHelp()</tt> method.</p>
	 * 
	 * @param        args Arguments sent using the command line when the program is launched.
	 */	
	public static void main(String args[])
	{
		// creating a new log file
		log = new LogWriter();
		
		// if no argument was given, displays help
		if (args.length == 0)
			displayHelp();
		
		// if at least one argument is given, it depend on its value:
		else if (args.length > 0)
		{
			// displaying help
			if ( args[0].equals("--help") )
				displayHelp();

			// creating the .sql file
			else if ( args[0].equals("--sql-file") )
				sqlFile();
			
			// creating a new splitsort, but checking the config file and the connection with the database first.
			else if ( args[0].equals("--new") )
			{
				if (args.length <= 1)
					displayHelp();	// if no password was given, displays help
				else if (Interpreter.checkFile(log))
					if ( MySQLbase.testConnection(log, args[1]) )
						generateJob(args[1]);	// if all the tests were correct, a job is generated
				else
					System.out.println("\nERROR : a problem was encountered. Check logs for more details.");	
			}
			
			// launching one iteration, but checking the config file and the connection with the database first.
			else if ( args[0].equals("--iteration") )
			{
				if (args.length <= 1)
					displayHelp();	// if the password was not given, displays help
				else if (Interpreter.checkFile(log))
					if ( MySQLbase.testConnection(log, args[1]) )	// if the tests were correct:
					{
						useResults(args[1]);	// results are used to finish a quicksort iteration
						comp.display();
						if (keepGoing)
							generateJob(args[1]);	// a new one is launched and a new job is created if it is not finished
					}
				else
					System.out.println("\nERROR : a problem was encountered. Check logs for more details.");
				else
					System.out.println("\nERROR : a problem was encountered. Check logs for more details.");
			}
			 
			
			// checking...
			else if (args[0].equals("--check"))
			{
				// ... nothing (not enough arguments)
				if (args.length == 1)
					displayHelp();
				
				// ... the content of the configuration file
				if (args[1].equals("config"))
				{
					if (Interpreter.checkFile(log))
						System.out.println("\nThe config.xml file contains all the information necessary.\n\n");
					else
						System.out.println("\nA problem was encountered (information missing)." +
								"Check log for more information\n");
				}
				
				// ... the connection with the database
				else if (args[1].equals("sql-db"))
				{
					if (args.length <= 2)
						displayHelp();	// if the password was not given, displays help
					else if ( MySQLbase.testConnection(log, args[2]) )
						System.out.println("connection with the database successfully created");
					else
						System.out.println("A problem was encountered while connecting with the database." +
							"Check log for more information\n");
				}

				// ... the creation of a HIT
				else if (args[1].equals("newjob"))
				{
					config = Interpreter.getConfig() ;
					CrowdManager test = new CrowdManager(Integer.parseInt( config.get("cfJobNumber") ), config.get("cfKey"),
							config.get("antechamberUrl"), null, null, Interpreter.getFields("CSVfield") );
					System.out.println(" new job identifier : " + test.createNewJob());
				}
				
				// ... the sending of a final e-mail when everything is finished
				else if (args[1].equals("final-email"))
				{
					config = Interpreter.getConfig();
					CrowdManager.thisIsTheEnd(config.get("antechamberUrl") + "main.php",
							"<?xml version='1.0' encoding='UTF-8'?>\n" +
							"<results>\n  <media id='99'>\n    <timecode>99</timecode>\n    <name>truc99.ogg</name>\n" +
							"    <axe1>-0.8181818181818181</axe1>\n  </media>\n</results>");
					System.out.println("Signal sent at " + config.get("antechamberUrl") + "main.php\n");
				}
				
				// ... nothing because the argument given was invalid: displays help
				else
					displayHelp();
			}
			
			// displaying help otherwise
			else
				displayHelp();
		}
		
		// writing the date and time of end in the log file.
		log.finish();
		System.out.println();
	}


	/**
	 * Saves the results, either intermediary ones if <tt>finished</tt> is <tt>false</tt> or final ones if it is
	 * <tt>true</tt>.
	 * 
	 * <p>If the <tt>finished</tt> parameter is <tt>false</tt>, then the current state of the splitsort iswritten in 
	 * as much <tt>< axis>.xml</tt> files as there are axes, with <tt>< axis></tt>'s value going through all the axes
	 * considered. Such results are formatted so as to be easily retrieved during the next iteration.</p>
	 * <p>If <tt>finished</tt> is true, final results are saved in the <tt>../results.xml</tt> file. It contains all the
	 * information in the <tt>media.xml</tt> file plus the coordinates of each one of them along each axis. When the file
	 * is written, it is sent to the Antechamber using . Such
	 * call should only be made when the sort is done.</p> 
	 * 
	 * @param	finished	If true, write definitive results ; if false, write temporary ones.
	 * 
	 * @see Interpreter#generateIntermediaryXMLs(String, List)
	 * @see Interpreter#generateFinalXML(String, Map)
	 * @see crowdUser.CrowdManager#thisIsTheEnd(String, String)	
	 */
	public static void save(Boolean finished)
	{
		if (finished)
		{
			Map<String,Integer[]> xmlParameters = new Hashtable<String,Integer[]>();
			// writing results in the shell, formatted like a python dictionnary to make it easier to check it.
			System.out.println("This is for '../data/CM-stub/stub.py check' if you where testing CPS:");
			System.out.print("{ ");
			for ( Iterator<String> str = comp.getSplitSorts().keySet().iterator(); str.hasNext(); )
			{  
				String axis = str.next();
				Integer[] res = comp.getSplitSorts().get(axis).getFinalResult();
				System.out.print(" '" + axis + "' : [");
				for (int i=0; i<res.length; i++)
					System.out.print(res[i]+", ");
				System.out.print("] ,  ");
				xmlParameters.put(axis, res);
			}
			System.out.print("}\n\n");
			// saving results in the ../data/results.xml file
			String results = inter.generateFinalXML("../results.xml",xmlParameters);
			log.append("[DONE] writing results in results.xml");
			// sending final email
			CrowdManager.thisIsTheEnd(config.get("antechamberUrl") + "main.php",results);
			log.append("[DONE] sending end notification e-mail.");
			System.out.println("\n\nThis splitsort is finished! You can now get the results in the data/results.xml file.");
		}
		else
			for (String axis : comp.getSplitSorts().keySet())
			{
				log.append("[DONE] writing temporary state in " + axis + ".xml");
				inter.generateIntermediaryXMLs("../data/previously/" + axis + ".xml",comp.getCurrentState().get(axis));
			}
	}



	/**
	 * Initializes the classes necessary for a splitsort iteration and checks whether the sort must go on or not.
	 * 
	 * <p>In this method, several classes are initialized using the content of the <tt>media.xml</tt> and 
	 * <tt>config.xml</tt> files. They correspond to all of the MainClass attribute.</p>
	 * <p>It also checks whether the sort should go on using the {@link Comparator#notFinished( )} method.</p>
	 * 
	 * @param	dbPasswd	The password of the MySQL database used.
	 * 
	 * @return	<tt>true</tt> if the sort should continue, <tt>false</tt> if it is already over.
	 */
	public static void initialize(String dbPasswd)
	{
		config = Interpreter.getConfig() ;
		log.append("[DONE] reading configuration file");
		
		mediaBdd = new MediaBase(config.get("mediaPath"));
		log.append("[DONE] reading media to be sorted");
		
		inter = new Interpreter(mediaBdd);
		log.append("[DONE] initialisation of the Interpreter");
		
		Integer drop = Integer.parseInt(config.get("drop"));		// the size of the arrays under which they are not sorted
		log.append("[DONE] array to be sorted created\n   Note : the sort will stop for arrays having" +
				" a smaller than " + drop + " size");
		
		Map<String,SplitSort> splitsortMap = new HashMap<String,SplitSort>();
		List<String> axes = Interpreter.getAxes();
		
		keepGoing = false;	// "default" value
		for (String axis : axes)
		{
			allMedia = inter.getCurrentState(log,axis);
			SplitSort sorter = new SplitSort(allMedia,drop);
			splitsortMap.put(axis,sorter);
			if (sorter.notFinished())
				keepGoing = true;
		}
		
		// If the sort must go on :
		if (keepGoing)
		{
			hitBdd = new MySQLbase("jdbc:mysql://localhost:3306/" + config.get("sqlBase") , config.get("sqlUser"),
					dbPasswd, Interpreter.getFields("SQLfield"), Interpreter.getFields("SQLtype"));
			log.append("[DONE] connexion to mySQL database");
			
			comp = new Comparator(splitsortMap,hitBdd);
			log.append("[DONE] initialisation of the comparator");
			
			theBoss = new CrowdManager(Integer.parseInt( config.get("cfJobNumber") ), config.get("cfKey"),
					config.get("antechamberUrl"), hitBdd, mediaBdd, Interpreter.getFields("CSVfield") );
			log.append("[DONE] initialisation of the CrowdManager");
		}
		// Otherwise, final results are written without going further.
		else
		{
			comp = new Comparator(splitsortMap);
			log.append("[DONE] initialisation of the comparator.\n   Note: No connexion to the My SQL database created.");
			save(true);
		}
	}


	/**
	 * Performs the first half of a quicksort iteration, generates the comparisons it requires and submits them in a new
	 * job at CrowdFlower.
	 *  
	 * <ol>
	 * 		<li>First, this class' attributes are set using the <tt>initialise</tt> method.</li>
	 * 		<li>The Comparator instance performs the first step of a quicksort iteration and thus generates a list of
	 * 			unknown comparisons.</li>
	 * 		<li>These comparisons are sent to a CrowdManager instance.</li>
	 * 		<li>If it is necessary (i.e, if there ARE unknown comparisons), the CrowdManager instance then generates the
	 * 			 HITs and uploads them to the right job at CrowdFlower.</li>
	 * </ol>
	 * 
	 * @param	dbPasswd	The password of the database matching the user specified in the config.xml file.
	 */
	public static void generateJob(String dbPasswd)
	{
		log.append("\n___________ generating Jobs  ___________");
		
		// step 1: initialization
		initialize(dbPasswd);
		if (keepGoing)
		{
			
			// step 2: first step of the quicksort
			comp.sortIteration();
			log.append("[DONE] first step of the quicksort");
			
			// step 3: sending unknown comparisons to a CrowdManager instance
			comp.demandNewComparisons(theBoss.getGreater(),theBoss.getSmaller(),theBoss.getAxes());
			log.append("[DONE] asking for new comparisons\n   Note : " +
	    		  theBoss.getGreater().size() + " new comparisons required") ;
			
			// step 4: uploading a new job if necessary
			if (theBoss.getGreater().size() >= 1)
			{
				Integer jobId = theBoss.createNewJob();
				Interpreter.setLastJob(jobId);
				log.append("[DONE] creating a new job at CrowdFlower (id=" + jobId + ")");
				theBoss.uploadHIT(jobId);
				log.append("[DONE] uploading HIT\n   ... Now waiting for HIT results.");
			}
			else
			{
				log.append("   Note : all comparisons are already known : no HIT");
				useResults(dbPasswd);
			}
		}
	}


	/**
	 * Human made results are downloaded, interpreted, ameliorated and put in the {@link myDataBases.MySQLbase} instance
	 * <tt>hitBdd</tt> allowing the second half of a quicksort iteration to be performed.
	 *  
	 * <ol>
	 * 		<li>The MainClass' attributes are initialized.</li>
	 * 		<li>Results are read and ameliorated using code from P. Ipeirotis and is students' program,
	 * 			<a href="http://code.google.com/p/get-another-label/">	get another label</a>.</li>
	 * 		<li>The Comparator instance performs the last step of the quicksort.</li>
	 * 		<li>A dump of the DBcps database is created to make sure that, even if this program fails, it won't be necessary to
	 * 			submit again already sent HITs.</li>
	 * 		<li>The results are saved along each axis in an xml file.</li>
	 * </ol>
	 * 
	 * 
	 */
	public static void useResults(String dbPasswd)
	{
		log.append("\n___________ Using Results  ___________");
		
		// step 1: initialization
		initialize(dbPasswd);
		if (keepGoing)
		{
			comp.sortIteration();	// Each splitsort recreates its comparisons in order to be able to understand the results
			String path2results = theBoss.getHITresults( Interpreter.getLastJob() );
			
			// step 2 : after the Job is done, the results are read and ameliorated.
			theBoss.readHITresults( path2results );
			log.append("[DONE] reading HIT results");
			
			// step 3 : last step of the quicksort
			comp.endIteration();
			log.append("[DONE] last step of the quicksort");
			
			// step 4 : generating a dump of the database
			hitBdd.generateDump();
			log.append("[DONE] dump of the MySQL database");
			
			// step 5 : saving current results
			save(false);
		}
	}



	/**
	 * Creates a .sql file allowing the creation of the database and the tables necessary for CPS.
	 * 
	 * <p>It contains the information necessary to create the <tt>hit</tt> and <tt>results</tt> table, each one having
	 * the fields necessary for the CPS to work as well as those added in the config.xml file. Each table is creating
	 * in a different way:</p>
	 * <ul>
	 * 		<li><tt>hit</tt>'s content depends completely of the content of the config.xml file ('hardDataFields' node)</li>
	 * 		<li><tt>results</tt> contains axis, idMedia1, idMedia2, greater, nameMedia1, nameMedia2 and what is in the
	 * 'resultsFields' node of the config.xml file.</li>
	 * </ul>
	 * <p>The file generated is <tt>../data/mysql/createCPStables.sql</tt>.</p>
	 */
	public static void sqlFile()
	{
		String user = Interpreter.getConfig().get("sqlUser");
		String fileContent = "CREATE DATABASE DBcps ;\n" +
				"GRANT DELETE,DROP,SELECT,INSERT,LOCK TABLES,UPDATE,CREATE ON DBcps.* TO '" +
				user + "'@'localhost' ;\n" +
				"USE DBcps ;\n" +
				"CREATE TABLE results (axis VARCHAR(40), idMedia1 INT, idMedia2 INT, greater TINYINT," +
				" nameMedia1 VARCHAR(300), nameMedia2 VARCHAR(300) ) ;\n" +
				"CREATE TABLE hit (";
		List<String> fields = Interpreter.getFields("SQLfield");
		List<String> types = Interpreter.getFields("SQLtype");
		for (int i=0; i<fields.size()-1; i++)
			fileContent += fields.get(i) +" "+ types.get(i) +", ";
		fileContent += fields.get(fields.size()-1) +" "+ types.get(fields.size()-1) +" ) ;";
		try
		{
			BufferedWriter sqlFile = new BufferedWriter
		  			(new FileWriter("../data/mysql/createCPStables.sql",false));
			sqlFile.write(fileContent);
			sqlFile.newLine();
			sqlFile.close();
		}
		catch (Exception e) { e.printStackTrace(); }
		System.out.println("File created : data/mysql/createCPStables.sql");
	}


	/**
	 * Displays this program's help and exit.
	 * 
	 * <p>The appearance of the help displayed is copied from typical UNIX man pages.</p>
	 */
	public static void displayHelp()
	{
		System.out.println("\n\n\n =========================== CROWD-POWERED-SORT =========================== ");
		System.out.println();
		System.out.println("Sorts media along specified axis and returns the results in a file (xml format).");
		System.out.println("A CrowdFlower (<http://crowdflower.com>) account is necessary.");
		System.out.println();
		System.out.println(" OPTIONS");
		System.out.println();
		System.out.println("	--new 'DBpasswd'	Launches the splitsort of the media contained in the ");
		System.out.println("				data/media.txt file or in the data/media folder (this must be ");
		System.out.println("				precised in the config.xml file). 'DBpasswd' is the password of");
		System.out.println("				the database. In order for the sort to continue, you must use");
		System.out.println("				this program again, but with the '--iteration' option when you");
		System.out.println("				receive an e-mail.");
		System.out.println();
		System.out.println("	--iteration 'DBpasswd'	Works just like '--sort' but launches a unique ");
		System.out.println("				splitsort iteration. Results are stored as xml files");
		System.out.println("				corresponding to each axis in the '../data/tmp' folder.");
		System.out.println();
		System.out.println("	--sql-file	creates a .sql file which, uploaded with the mysql root account,");
		System.out.println("			will create the database DBcps and the tables necessary for CPS");
		System.out.println("			to work, i.e 'hit' and 'results'.");
		System.out.println();
		System.out.println("	--check=[OPTION]	this option can have several value :");
		System.out.println("		config		checks the content of the config.xml file.");
		System.out.println("		sql-db 'DBpasswd' checks the connection with the MySQL database.");
		System.out.println("		newjob	attempts to create a new job by copying the one whose id");
		System.out.println("				is specified in the config.xml file");
		System.out.println("		final-email	Asks the antechamber to send a final notification email");
		System.out.println("				spotting the end of an imaginary splitsort. Fake results are");
		System.out.println("				attached in order to check this as well.");
		System.out.println();
		System.out.println("	--help displays this help and exit");
		System.out.println("\n\n\n");
	}
}
