package sorter;

import java.util.*;
import myDataBases.MySQLbase;


/**
 * Uses several instances of {@link SplitSort}, one for each axis considered, to sort media. 
 * 
 * <p>The aim of this class is to manage a SplitSort instance for each axis considered and to ask for all comparisons
 * necessary at a time, <em>i.e</em> to "group" them in order to limit the number of jobs uploaded as well as allowing a 
 * treatment of the results requiring a lot of data to be processed at the same time.</p>
 * 
 * <p>A lot of its methods are simply calls to their equivalent for each SplitSort instance in the 
 * <tt>{@link Comparator.sorters}</tt> attribute.</p>
 * 
 * @author Leo Perrin (perrin.leo@gmail.com)
 */
public class Comparator
{

	//
	// Fields
	//

	/**
	 * This map must have the following structure : [ key = <tt>axis considered</tt>,
	 * element = <tt>instance of SplitSort in charge of sorting media along this axis</tt> ].
	 */
	private Map<String,SplitSort> sorters;
	/**
	 * The {@link myDataBases.MySQLbase} database from which comparisons will be retrieved.
	 */
	private MySQLbase sqlBase;
	
	
	//
	// Constructors
	//

	/**
	 * Creates a Comparator instance using the SplitSorts contained in the <tt>sorter</tt> map and links it with a
	 * <tt>{@link myDataBases.MySQLbase}</tt> instance.
	 * 
	 * <p>Both of the attributes of this class are set when this constructor is called.</p>
	 * 
	 * @param        sorter This map should have the following structure : [ key = <tt>axis considered</tt>,
	 * element = <tt>instance of SplitSort in charge of sorting media along this axis</tt> ].
	 * @param        bdd The database that will be used as the "sqlBase" attribute.
	 */
	public Comparator( Map<String,SplitSort> sorter, myDataBases.MySQLbase bdd )
	{
		this.sorters = sorter;
		this.sqlBase = bdd;
	}


	/**
	 * Creates a Comparator instance using the SplitSorts contained in the <tt>sorter</tt> map.
	 * 
	 * <p>This constructor sets the value the <tt>sorter</tt> attribute, but <strong>not</strong> that of the MySQLbase !
	 * Be careful therefore not to do anything requiring it with such an instance.</p>
	 * 
	 * @param        sorter This map should have the following structure : [ key = <tt>axis considered</tt>,
	 * element = <tt>instance of SplitSort in charge of sorting media along this axis</tt> ].
	 */
	public Comparator( Map<String,SplitSort> sorter )
	{
		this.sorters = sorter;
	}	
	
	//
	// Methods
	//
	
	/**
	 * Returns <tt>true</tt> while the sorting is not over.
	 * 
	 * <p>It equals false only if {@link SplitSort#notFinished( )} is false for all the SplitSorts in <tt>sorters</tt>,
	 * in which case quicksort iterations must stop.</p>
	 * 
	 * @return	<tt>true</tt> if the sort must continue, <tt>false</tt> otherwise (case in which it is over).
	 */
	public Boolean notFinished()
	{
		Boolean result = false;
		for (Iterator<String> str = this.sorters.keySet().iterator(); str.hasNext();)
			if (this.sorters.get(str.next()).notFinished() == true)
			{
				result = true;
				break;
			}
		return result;
	}

  
	/**
	 * Performs the first part of a quicksort iteration.
	 * 
	 * <p>All of the SplitSort instances contained in "sorters" are asked to perform the <strong>first</strong> step of a
	 * quicksort iteration.</p>
	 */
	public void sortIteration( )
	{
		for (Iterator<String> str = this.sorters.keySet().iterator(); str.hasNext();)
		{
			String axis = str.next();
			if (this.sorters.get(axis).notFinished())
				this.sorters.get(axis).iteration();
		}
	}


	/**
	 * Performs the last part of a quicksort iteration.
	 * 
	 * <p>All of the SplitSort instances contained in <tt>sorters</tt> are asked to perform the <strong>last</strong> step
	 * of a quicksort iteration.</p>
	 */
	public void endIteration(  )
	{
		List<Boolean> comparisons = new ArrayList<Boolean>();
		List<Integer> greater = new ArrayList<Integer>();
		List<Integer> smaller = new ArrayList<Integer>();  
		for (String axis : this.sorters.keySet() )
		{
			comparisons.clear();
			greater = this.sorters.get(axis).getGreater();
			smaller = this.sorters.get(axis).getSmaller();
			for (int i=0; i<greater.size(); i++)
				if (greater.get(i) == smaller.get(i))
					comparisons.add(false);
				else
					comparisons.add( this.sqlBase.comparisonResult(greater.get(i),smaller.get(i),axis) );
			this.sorters.get(axis).endIteration(comparisons);
		}
	}



	/**
	 * Asks for unknown comparisons by modifying the value of its parameters <tt>greater</tt>, <tt>smaller</tt> and
	 * <tt>axes</tt>.
	 * 
	 * <p>Returns in the parameters <tt>greater</tt> and <tt>smaller</tt> the identifiers of the elements that require a
	 * comparison and are alleged to be respectively greater and smaller and, in <tt>axes</tt>, the axes along which the
	 * comparison must be performed.</p>
	 * 
	 * @param     greater All the media to be compared to the pivots in the several pending quicksorts.
	 * @param     smaller The pivots corresponding to the several pending quicksorts.
	 * @param		axes	The axes along which comparisons must be performed.
	 */
	public void demandNewComparisons( List<Integer> greater, List<Integer> smaller, List<String> axes)
	{
		List<Integer> sortGreater = new ArrayList<Integer>();
		List<Integer> sortSmaller = new ArrayList<Integer>();
		greater.clear();
		smaller.clear();
		axes.clear();
		
		for (Iterator<String> str = this.sorters.keySet().iterator(); str.hasNext();)
		{
			String emotion = str.next();
			sortGreater = this.sorters.get(emotion).getGreater();
			sortSmaller = this.sorters.get(emotion).getSmaller();
			for (int i=0; i<sortGreater.size(); i++)
				if ( sortGreater.get(i) != sortSmaller.get(i) )
					if (this.sqlBase.alreadyDone(sortGreater.get(i), sortSmaller.get(i),emotion) == false)
					{
						greater.add(sortGreater.get(i));
						smaller.add(sortSmaller.get(i));
						axes.add(emotion);
					}
		}
	}
	

	/**
	 * Displays the current state of the sort using <tt>{@link SplitSort#display()}</tt>.
	 */
	public void display()
	{
		for (String axis : this.sorters.keySet())
		{
			System.out.println("\n"+axis);
			this.sorters.get(axis).display();
		}
	}
	//
	// Accessor methods
	//

	/**
	 * Returns the <tt>sorters</tt> attribute.
	 * 
	 * @return	A Map<String,SplitSort> containing the SplitSort instance used as well as the axis associated with each
	 * one of them.
	 */
	public Map<String,SplitSort> getSplitSorts()
	{  return this.sorters;  }

	
	/**
	 * Returns the current state of the sort, i.e a Map indexed by the axes and containing the <tt>sorted</tt> attribute
	 * of each SplitSort contained in <tt>sorters</tt>.
	 * 
	 *  @return		A map containing the current state of the sort.
	 */
	public Map<String, List<Integer[]> > getCurrentState()
	{
		Map<String, List<Integer[]>> state = new Hashtable<String, List<Integer[]>>() ;
		for(String axis : this.sorters.keySet())
			state.put(axis, this.sorters.get(axis).getSorted());
		return state;
	}
}
	