package sorter;

import java.util.*;


/**
 * Implements the splitsort, a quicksort in which comparisons are completely separated from the rest of the algorithm.
 * 
 * <p>This class provides an implementation of the SplitSort : if given an array of integers, it will return, for each
 * quicksort iteration, the comparisons that are necessary to continue. It allows the first and last step of a quicksort
 * iteration (i.e generating comparison and splitting each array into three smaller ones) to be performed separately so
 * as the comparisons to be gathered with those of other SplitSort instances (this is roughly the role of the
 * {@link Comparator} class).</p>
 * <p>In order not to submit to many comparisons, it is possible to stop quicksort iterations for an array when its size is
 * smaller than <tt>minimalSize</tt>. Since each comparison has a cost, it can be interesting to limit their number.</p>
 * <p>An array containing the sorted elements is generated in the end.</p>
 * 
 * @author Leo Perrin (perrin.leo@gmail.com)
 */
public class SplitSort
{

	//
	// Fields
	//

	/**
	 * Contains all the elements that are to be compared with their corresponding pivot.
	 * <p>These inside this list are alleged to be greater than their counterpart in the <tt>smaller</tt> list. There is
	 * no mathematical basis whatsoever for this statement, it simply allows the result of a comparison to be easily
	 * understood : if <tt>comparisons.get(k)</tt> is <tt>true</tt>, then <tt>greater.get(k)</tt> IS greater than
	 * <tt>smaller.get(k)</tt>. Otherwise, it is the contrary.
	 */
	private List<Integer> greater;
	/**
	 * Contains all the pivots used in the current iteration of the quicksort.
	 * <p>The media whose ID are inside this list are alleged to be smaller than their counterpart in the <tt>greater</tt>
	 * list (this statement is explained in the description of <tt>greater</tt>).</p>
	 */
	private List<Integer> smaller;
	/**
	 * Contains the elements to be sorted in a list of arrays.
	 * <p>Each array will go through an iteration of the quicksort. Using this list of arrays allows an iterative
	 * implementation of the quicksort rather than the "classical" recursive one.</p>
	 */
	private List<Integer[]> sorted;
	/**
	 * Contains <tt>true</tt> at the rank <tt>k</tt> if the <tt>k</tt>-th element of "greater" <strong>is</strong>
	 * actually greater than that of <tt>smaller</tt>: <tt>this.comparisons.get(k) = ( this.greater.get(k) > 
	 * this.smaller.get(k) )</tt>, where <tt>'>'</tt> does <strong>not</strong> correspond to a comparison between integers
	 * but to the one given by a Comparator instance.
	 */
	private List<Boolean> comparisons;
	/**
	 * The number of elements contained in the initial list of arrays to sort.
	 */
	private Integer numberOfElements;
	/**
	 * An array of a length smaller than <tt>minimalSize</tt> will <strong>not</strong> be sorted.
	 */
	private Integer minimalSize;

	//
	// Constructor
	//

	/**
	 * Creates an instance of SplitSort and sets all of its attribute.
	 * 
	 * <p>Creates an instance of SplitSort and sets the value of three of its attributes :
	 * <ul>
	 * 	<li><tt>{@link SplitSort#sorted}</tt>
	 * 	<li><tt>{@link SplitSort#numberOfElements}</tt>
	 * 	<li><tt>{@link SplitSort#minimalSize}</tt>
	 * </ul>
	 * </p>
	 * 
	 * @param	init	The list of array containing all the elements to be sorted as well as the current state of the sort.
	 * @param   min		The value to be assigned to <tt>minimalSize</tt>.
	 */
	public SplitSort( List<Integer[]> init, Integer min )
	{
		this.sorted = new ArrayList<Integer[]>();
		this.greater = new ArrayList<Integer>();
		this.smaller= new ArrayList<Integer>();
		this.comparisons = new ArrayList<Boolean>();
		this.minimalSize = min;
		this.numberOfElements = 0;
		for (Integer[] array : init)
		{
			this.sorted.add(array.clone());
			this.numberOfElements += array.length ;
		}
	}
	
	//
	// Methods
	//


	/**
	 * Returns true if the sorting of the elements in this instance of SplitSort is not finished.
	 * 
	 * <p>To determine whether the sort is over or not, it calculates the number (<tt>count</tt>) of elements contained in 
	 * all arrays containing less than <tt>minimalSize</tt> elements. It stops when a bigger array is found, i.e
	 * (<tt>array.length > minimalSize)</tt>. If <tt>count</tt> equals the number of elements to be sorted, then it means
	 * all elements have been sorted : it is finished ! Otherwise, well, it's not.</p>
	 * 
	 * @return	 <tt>true</tt> if the sort is over, <tt>false</tt> otherwise.
	 */
	public Boolean notFinished()
	{
		int count = 0;
		for (Integer[] tab : this.sorted)
			if (tab.length > this.minimalSize)
				break;
			else
				count+=tab.length;
		return (count != this.numberOfElements);
	}
	

	/**
	 * Performs the first part of an iteration of the quicksort on all the arrays contained in <tt>sorted</tt>.
	 * 
	 * <p>This means choosing a pivot in each array contained in <tt>sorted</tt> and generating all of the comparisons
	 * necessary to go further. This methods :
	 * <ol>
	 * 		<li>Clears the content of its attributes <tt>comparisons</tt>, <tt>greater</tt> and <tt>smaller</tt> to avoid
	 * 		former results to interfere.</li>
	 *	 	<li>Calls <tt>{@link SplitSort#addComparisons(Integer[])}</tt> on each array contained in <tt>sorted</tt>.</li>
	 * </ol></p>
	 */
	public void iteration()
	{
		this.comparisons.clear();
		this.greater.clear();
		this.smaller.clear();
		for ( Integer[] tab : this.sorted)
			if (tab.length > this.minimalSize)
				addComparisons(tab);
	}


	/**
	 * Fills the <tt>greater</tt> and <tt>smaller</tt> arrays with the content of the <tt>tab</tt>parameter.
	 * 
	 * <p>The comparisons necessary to perform a quicksort on the array "tab" are added to the attributes <tt>greater</tt>
	 * and <tt>smaller</tt>. <tt>greater</tt> contains all the element of tab while <tt>smaller</tt> is of the same size
	 * but contains only the pivot that will be used for the quicksort, i.e the first element of <tt>tab</tt>.</p>
	 * 
	 * @param        tab The array that will go through an iteration of quicksort (using a pivot).
	 */
	private void addComparisons( Integer[] tab )
	{
		for (int i=0; i<tab.length; i++)
		{
			this.greater.add(tab[i]);
			this.smaller.add(tab[0]);
		}
	}



	/**
	 * Performs the end of the quicksort iteration on all of the arrays contained in the attribute list
	 * <tt>{@link SplitSort#sorted}</tt>.
	 * 
	 * <p>The end of the quicksort iteration corresponds to the splitting of all arrays contained in <tt>sorted</tt> in at
	 * least two parts. See the <tt>{@link SplitSort#insertion(int)}</tt> method's description for more details.</p>
	 * 
	 * @param     comparisons	A list of Boolean containing the results of the comparisons. Most of part this program's
	 * 							only purpose is to create this list...
	 * @see SplitSort#insertion(int)
	 */
	public void endIteration( List<Boolean> comparisons )
	{
		int i=0;
		this.comparisons = comparisons;
		while (i<this.sorted.size())
			if ( this.sorted.get(i).length > this.minimalSize)
				i+=insertion(i);
			else
				i+=1;
	}


	/**
	 * Performs the end of a quicksort iteration on the array at the <tt>rank</tt>-th position in the <tt>sorter</tt>
	 * list and returns the number of arrays that have been inserted in it.
  	 *
     * <p>This is done by splitting this array in three arrays :
     * <ul>
     * 		<li>One that contains all the elements greater than the pivot.</li>
     * 		<li>One that contains only the pivot.</li>
     * 		<li>One that contains all the elements smaller than the pivot.</li>
     * </ul>
     * If the pivot is actually the smallest or the greatest element of this array, only two arrays are created. The "old"
     * array is removed from <tt>sorted</tt>.
     * </p>
     * 
     * @param	rank 	The array at the <tt>rank</tt> rank in the list <tt>sorted</tt> will be split in several parts.
     * 
     * @return  The number of arrays actually inserted.
     */
	private int insertion( int rank )
	{
		// initialisations
		Integer[] tab = this.sorted.get(rank);
		int pivot = tab[0];
		List<Integer> small = new ArrayList<Integer>() ;
		List<Integer> great = new ArrayList<Integer>() ;
		int pivotIndex = this.smaller.indexOf(pivot);
		
		// creation of the arrays to add
		for (int j=0; j<tab.length; j++)
		{
			int g = this.greater.get(pivotIndex+j);
			int s = this.smaller.get(pivotIndex+j);
			if (this.comparisons.get(pivotIndex+j))
			{
				if (s!=pivot) small.add(s);
				if (g!=pivot) great.add(g);
			}
			else
			{
				if (g!=pivot) small.add(g);
				if (s!=pivot) great.add(s);
			}
		}
		Integer[] great_array = great.toArray(new Integer[0]);
		Integer[] array_pivot =  { pivot };
		Integer[] small_array = small.toArray(new Integer[0]);
		
		// removing and insertion of the right arrays
		this.sorted.remove(rank);
		int ins = 1;	// number of arrays inserted (at least the pivot)
		if (great_array.length>0)
		{
			this.sorted.add(rank,great_array);
			ins++;
		}
		this.sorted.add(rank,array_pivot);
		if (small_array.length>0)
		{
			this.sorted.add(rank,small_array);
			ins++;
		}
		
		// the number of arrays inserted is returned
		return ins;
	}


	/**
	 * Returns the content of the <tt>sorted</tt> attribute transferred in a simple array.
	 * 
	 * <p>For each array in the <tt>sorted</tt> list, its elements are appended to a <tt>result</tt> list of
	 * <tt>Integer</tt> (initialized empty). Two loops are used : one iterates on the list and another iterates on the
	 * elements of each array.</p>
	 * 
	 * @return 	An array of <tt>Integer</tt> containing the identifiers of the media sorted.
	 */
	public Integer[] getFinalResult(  )
	{
		List<Integer> result = new ArrayList<Integer>();
		for(int i=0; i<this.sorted.size(); i++)
			for (int j=0; j<this.sorted.get(i).length; j++)
				result.add(this.sorted.get(i)[j]);
		return result.toArray(new Integer[0]);
	}


	/**
	 * Prints the content of <tt>{@link SplitSort#sorted}</tt> on the screen, using brackets [] to separate each array.
	 * 
	 * <p>This method is very useful when debugging to see how the sort evolves, but rather useless otherwise. The
	 * <tt>{@link Comparator#display()}</tt> method that uses this <tt>SplitSort.display()</tt> should be commented most
	 * of the time.</p> 
	 */
	public void display(  )
	{
		for (Integer[] tab : this.sorted)
		{
			System.out.print(" [ ");
			for (Integer i : tab)
			{
				System.out.print(i);
				System.out.print(" ");
			}
			System.out.print("] ");
		}
	}


  
	//
	// Accessor methods
	//


	/**
	 * Get the value of <tt>greater</tt>.
	 * @return the value of <tt>greater</tt>
	 */
	public List<Integer> getGreater ( )
	{ return this.greater; }
	
	
	/**
	 * Get the value of <tt>smaller</tt>.
	 * @return the value of <tt>smaller</tt>
	 */
	public List<Integer> getSmaller ( )
  	{ return this.smaller; }

	
	/**
	 * Get the content of <tt>sorted</tt>.
	 * @return	the value of <tt>sorted</tt>
	 */
	public List<Integer[]> getSorted()
	{ return this.sorted; }

}
